/* 

  TIM.2- Things In Motion, version 2

  TIM (Things In Motion) is a set of related programs whose goal is to
  facilitate the visualization of time-series positional data.

  The data to be displayed can be either read from an existing file, or they
  can be generated by a simulation process and asynchronously displayed by a
  visualization process as the data are being generated. The visualization
  and simulation processes do not need to run on the same system; this
  allows the simulation process to be run on a large remote
  "number-cruncher", while the visualization can be done on a local system.

  TIM uses DIVERSE's dpf and dtk modules to allow the data to be viewed on
  systems ranging from a laptop to a multi-screen immersive display such as
  a CAVE(tm).

  Authors:
      John Kelso, Steve Satterfield
      Scientific Applications & Visualization Group
           SAVG web page: http://math.nist.gov/mcsd/savg/
      Mathematical & Computational Science Division
           MCSD web page: http://math.nist.gov/mcsd/
      National Institute of Standards and Technology
           NIST web page: http://www.nist.gov/


  For information about this software contact John Kelso, kelso@nist.gov

  "This software was developed at the National Institute of Standards and
  Technology by employees of the Federal Government in the course of their
  official duties. Pursuant to title 17 Section 105 of the United States
  Code this software is not subject to copyright protection and is in the
  public domain. This is an experimental system.  NIST assumes no
  responsibility whatsoever for its use by other parties, and makes no
  guarantees, expressed or implied, about its quality, reliability, or any
  other characteristic.

 We would appreciate acknowledgement if the software is used."

*/

#include <signal.h>
#include <stdio.h>
#include <stdlib.h>

#include <dtk.h> 

#include "getString.C"

/*
 reads positional data from stdin and writes it to dtk shared memory
 each position is an xyz translation, and a xyzw quaternion

*/

void usage(char *name) {
  fprintf(stderr,
	  "Usage: %s [--time t] [--number d] segname\n"
	  "where t is the time (float seconds) to wait between reads\n"
	  "      d is the number of positional elements in the "
	  "        shared memory segname\n\n"
	  " if t is omitted, 1/30th of second is used\n"
	  " if d is omitted, the size of segname is used if it exists\n"
	  " segname is created if it doesn't exist\n"
	  " if neither d is given and segname doesn't exist,\n"
	  " a size of one xyzxyzw is used for the new segment\n", name) ;
}

int running = 1;

void signal_catcher(int sig)
{
  running = 0;
}

int main(int argc, char **argv) {
  
  float t = (1.f/30.f) ;
  int d = 1 ;
  int d_specified = 0 ;
  char *segname = NULL ;
  
  char *name = strdup(argv[0]) ;
  
  if (argc == 0) {
    usage(name) ;
    exit(1) ;
  }
  
  int i = 1 ;
  while (i<argc) {
    if (segname) {
      fprintf(stderr, 
		 "Error: %s: parameter given after segment name: %s\n",
	      name,argv[i]) ;
      usage(name) ;
      return 1 ;
      
    } else if ((strstr("--time",argv[i]))!=NULL) {
      i++ ;
      if (i<argc) {
	if (1 != sscanf(argv[i],"%f",&t)) {
	  fprintf(stderr, 
		     "Error: %s:invalid time = %s\n",name, argv[i]) ;
	  usage(name) ;
	  return 1 ;
	} else if (t <= 0.0f) {
	  fprintf(stderr, 
		     "Error: %s: invalid time = %s\n",name, argv[i]) ;
	  usage(name) ;
	  return 1 ;
	} else {
	  i++ ;
	}
      } else {
	fprintf(stderr, 
		   "Error: %s: No --time parameter\n", name) ;
	usage(name) ;
	return 1 ;
      }
       
    } else if ((strstr("--number",argv[i]))!=NULL) {
      i++ ;
      if (i<argc) {
	if (1 != sscanf(argv[i],"%d",&d)) {
	  fprintf(stderr, 
		     "Error: %s: invalid number of items = %s\n",
		  name, argv[i]) ;
	  usage(name) ;
	  return 1 ;
	} else if (d <= 0) {
	  fprintf(stderr, 
		     "Error: %s: invalid number of items = %s\n",
		  name, argv[i]) ;
	  usage(name) ;
	  return 1 ;
	} else {
	  i++ ;
	}
      } else {
	fprintf(stderr, 
		   "Error: %s: no --number parameter\n", name) ;
	usage(name) ;
	return 1 ;
      }
      d_specified = 1 ;

    } else { // segname
      segname = strdup(argv[i]) ;
      i++ ;
    }
  }
  if (!segname) {
    fprintf(stderr, 
	       "Error: %s: no segment name parameter\n", name) ;
    usage(name) ;
    return 1 ;
  }
#if 0
  printf("time set to %f\n",t) ;
  printf("number set to %d\n",d) ;
  printf("segname = %s\n",segname) ;
#endif
  
  dtkSharedMem *shm ;
  //get shared memory, check size
  if (d_specified) {
    // use existing segment, or create if it doesn't exist
    shm = new dtkSharedMem(sizeof(float)*d*7, segname, DTK_GET) ;
    if (!shm || shm->isInvalid()) {
      fprintf(stderr,
		 "Error: %s: unable to create or get shared memory of specified size\n", name) ;
      return 1;
    } 
  } else {
    // use exisiting segment
    shm = new dtkSharedMem(segname) ;
    if (!shm || shm->isInvalid()) {
      fprintf(stderr,
		 "Error: %s: unable to get shared memory named %s\n",
	      name, segname) ;
      return 1;
    }
    int size = shm->getSize() ;
    if ((size%(sizeof(float)*7))) {
      fprintf(stderr,
		 "Error: %s: shared memory is not of correct size\n", name) ;
      return 1;
    } else {
      d = size/(sizeof(float)*7) ;
    }
  }
  
  float *pos = (float*) malloc(sizeof(float)*d*7) ;

  dtkTime tt;
  long double from_t = tt.get();
  
  signal(SIGTERM, signal_catcher);
  signal(SIGINT, signal_catcher);

  dtkRealtime_init(t);

  char *line = NULL ;
  // read/write data- 
  while (running) {
    long double to_t = tt.get();
    for (int i=0; i<d; i++) {
      if ((getString(&line))==EOF) {
	// modified to not quit, at EOF< jtk 9/03
	//running=0 ;
	continue ;
      }
      sscanf(line, "%f %f %f %f %f %f %f",
	    &(pos[i*7]), &(pos[i*7+1]), &(pos[i*7+2]), 
	    &(pos[i*7+3]), &(pos[i*7+4]), &(pos[i*7+5]), &(pos[i*7+6])) ;
#if 0
      printf("%d: %f %f %f %f %f %f %f\n", i,    
	     pos[i*7], pos[i*7+1], pos[i*7+2], 
	     pos[i*7+3], pos[i*7+4], pos[i*7+5], pos[i*7+6]) ;
#endif
    }
    
    // write to shared memory the local floats, swapping if necessary
    if(shm->differentByteOrder())
      dtk_swapBytes(pos, sizeof(float)*d*7);
    if (shm->write(pos)) {
      fprintf(stderr,
		 "Error: %s: unable to write shared memory segment %s\n",
	      name, segname) ;
      return 1 ;
    }
    from_t = to_t;
    dtkRealtime_wait();
  }

  dtkRealtime_close();
}
