/* 

  TIM.2- Things In Motion, version 2

  TIM (Things In Motion) is a set of related programs whose goal is to
  facilitate the visualization of time-series positional data.

  The data to be displayed can be either read from an existing file, or they
  can be generated by a simulation process and asynchronously displayed by a
  visualization process as the data are being generated. The visualization
  and simulation processes do not need to run on the same system; this
  allows the simulation process to be run on a large remote
  "number-cruncher", while the visualization can be done on a local system.

  TIM uses DIVERSE's dpf and dtk modules to allow the data to be viewed on
  systems ranging from a laptop to a multi-screen immersive display such as
  a CAVE(tm).

  Authors:
      John Kelso, Steve Satterfield
      Scientific Applications & Visualization Group
           SAVG web page: http://math.nist.gov/mcsd/savg/
      Mathematical & Computational Science Division
           MCSD web page: http://math.nist.gov/mcsd/
      National Institute of Standards and Technology
           NIST web page: http://www.nist.gov/


  For information about this software contact John Kelso, kelso@nist.gov

  "This software was developed at the National Institute of Standards and
  Technology by employees of the Federal Government in the course of their
  official duties. Pursuant to title 17 Section 105 of the United States
  Code this software is not subject to copyright protection and is in the
  public domain. This is an experimental system.  NIST assumes no
  responsibility whatsoever for its use by other parties, and makes no
  guarantees, expressed or implied, about its quality, reliability, or any
  other characteristic.

 We would appreciate acknowledgement if the software is used."

*/

/*
  Reads in objects and frames files and creates a libpfseq file.  

  A command line option specifies time between frames.  

  No DIVERSE or Performer code.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <dtk.h>

#include "normalize.C"
#include "getString.C"

void usage(char *name) {
  fprintf(stderr, 
	  "Usage: %s --objects objFile --frames frameFile --sequence seqFile\n"
	  "          [--time t] [--verbose] [--noclobber] \n"
	  "where t is the time between frames- if not specified it\n"
	  "      uses the libpfseq default\n"
	  "      objFile is an existing \"tim\" objects file\n"
	  "      frameFile is an existing \"tim\" frames file.  if frameFile\n"
	  "         is \"-\" stdin will be used.\n"
	  "      seqFile is the new \"seq\" file.  if seqFile\n"
	  "         is \"-\" stdout will be used.  No checking is done\n"
	  "         to see if seqFile already exists unless \"noclobber\"\n"
	  "         is set\n", name) ;
}

int main(int argc, char **argv) {

  char *name = strdup(argv[0]) ;
  int i; //why not?
  int setTime = 0 ;
  int setVerbose = 0 ;
  int setNoClobber = 0 ;
  float time = 0 ;
  FILE *frames = NULL ;
  FILE *objects = NULL ;
  FILE *sequence = NULL ;
  char *seqFile = NULL ;

  // process arguments
  i = 1 ;

  while (i<argc) {
    if ((strstr("--frames",argv[i]))!=NULL) {
      i++ ;
      if (i<argc) {
	if (!strcmp("-", argv[i]))
	  frames = stdin ;
	else {
	  if (! (frames = fopen(argv[i], "r"))) {
	    fprintf(stderr,
		    "Error: %s: Can\'t open frames file %s\n",name, argv[i]) ;
	    usage(name) ;
	    return 1 ;
	  }
	}
	i++ ;
      } else {
	fprintf(stderr,
		"Error: %s: No --frames parameter\n",name) ;
	usage(name) ;
	return 1 ;
      }
      
    } else if ((strstr("--objects",argv[i]))!=NULL) {
      i++ ;
      if (i<argc) {
	if (! (objects = fopen(argv[i], "r"))) {
	  fprintf(stderr,
		  "Error: %s: Can\'t open objects file %s\n",argv[i], name) ;
	  usage(name) ;
	  return 1 ;
	} else {
	  i++ ;
	}
      } else {
	fprintf(stderr,
		"Error: %s: No --objects parameter\n", name) ;
	usage(name) ;
	return 1 ;
      }
      
    } else if ((strstr("--sequence",argv[i]))!=NULL) {
      i++ ;
      if (i<argc) {
	seqFile = strdup(argv[i]) ;
	i++ ;
      } else {
	fprintf(stderr,
		"Error: %s: No --sequences parameter\n", name) ;
	usage(name) ;
	return 1 ;
      }
      
    } else if ((strstr("--time",argv[i]))!=NULL) {
      i++ ;
      if (i<argc) {
	if (1 != sscanf(argv[i],"%f",&time)) {
	  fprintf(stderr,
		     "Error: %s: Invalid time = %s\n",name, argv[i]) ;
	  usage(name) ;
	  return 1 ;
	} else if (time <= 0.f) {
	  fprintf(stderr,
		     "Error: %s: Invalid time = %s\n",name, argv[i]) ;
	  usage(name) ;
	  return 1 ;
	} else {
	  i++ ;
	  setTime = 1 ;
	}
      } else {
	fprintf(stderr,
		   "Error: %s: No --time parameter\n", name) ;
	usage(name) ;
	return 1 ;
      }

    } else if ((strstr("--verbose",argv[i]))!=NULL) {
      i++ ;
      setVerbose = 1 ;

    } else if ((strstr("--noclobber",argv[i]))!=NULL) {
      i++ ;
      setNoClobber = 1 ;

    } else {
      fprintf(stderr,
	      "Error: %s: Unknown option: %s\n", name, argv[i]) ;
      usage(name) ;
      return 1 ;
      
    }

  }
  
  int bomb = 0 ;
  if (!frames) {
    fprintf(stderr,
	    "Error: %s: No --frames parameter supplied\n", name) ;
    bomb = 1 ;
  }
  
  if (!objects) {
    fprintf(stderr,
	    "Error: %s: No --object parameter supplied\n", name) ;
    bomb = 1 ;
  }
  
  if (!seqFile) {
    fprintf(stderr,
	    "Error: %s: No --sequence parameter supplied\n", name) ;
    bomb = 1 ;
  }


  if ((setNoClobber) && (strcmp("-",seqFile))) {
    if (fopen(seqFile, "r")) {
      fprintf(stderr,
	      "Error: %s: Sequence file, \"%s\", already exists, "
	      "but --noclobber specified\n",name, seqFile) ;
      bomb = 1 ;
      return 1 ;
    }
  }
  
  if (!(bomb)) {
    if (!strcmp("-", seqFile))
      sequence = stdout ;
    else
      if (!(sequence = fopen(seqFile, "w"))) {
	fprintf(stderr,
		"Error: %s: Can\'t open sequences file for writing %s\n",
		name, argv[i]) ;
	bomb = 1 ;
      }
  }
  
  if (bomb) {
    usage(name) ;
    return 1 ;
  }
  
  // first read in objects file
  int nFiles = 0 ;
  char **files = NULL ;
  char *input = NULL ;

  while ((getString(&input, objects))!=EOF){
    if (normalize(input)) {
      nFiles++ ;
      files = (char **) realloc(files, nFiles*sizeof(char *)) ;
      files[nFiles-1] = strdup(input) ;
    }
  }
  if (setVerbose)
    fprintf(stderr, "%s: read in %d file names\n",name, nFiles) ;

  fprintf(sequence, "#Beginning of data\n") ;

  // write time data
  if (setTime)
    fprintf(sequence,"time %f\n",time) ;

  // read in frame data
  int nFrames = 0 ;
  float xt, yt, zt, xq, yq, zq, wq ;
  while ((getString(&input, frames))!=EOF){
    if (normalize(input)) {
      nFrames++ ;
      if (0 == ((nFrames-1)%nFiles)) {
	fprintf(sequence, "\n###########\n# frame %d\n", nFrames/nFiles + 1) ;
	fprintf(sequence, "frame\n") ;
      }
      if ((nFrames-1)/nFiles == 0) {
	fprintf(sequence, "\nobject %s\n",files[(nFrames-1)%nFiles]) ;
	fprintf(sequence, "objectname %s\n",files[(nFrames-1)%nFiles]) ;
      } else {
	fprintf(sequence, "\nuseobject %s\n",files[(nFrames-1)%nFiles]) ;
      }
      if (7 != (sscanf(input, "%f %f %f %f %f %f %f",
		       &xt, &yt, &zt, &xq, &yq, &zq, &wq))) {       
	fprintf(stderr,
		"Error: %s: non conformant line: %s\n", name, input) ;
	return 1 ;
      } else {
	// note that the sequence file is ordered w, x, y, z
	if (xq != 0.f || yq != 0.f || zq != 0.f) {
	  dtkMatrix m ;
	  float h, p, r ;
	  m.quat(xq, yq, zq, wq) ;
	  m.rotateHPR(&h, &p, &r) ;
	  fprintf(sequence, "#hpr for below quat = (%3.3f %3.3f %3.3f)\n",
		  h, p, r) ;
	  fprintf(sequence, "quat %f %f %f %f\n", wq, xq, yq, zq) ;
	} else {
	  fprintf(sequence, "#zero quaternion, just use identity\n") ;
	  fprintf(sequence, "quat %f %f %f %f\n", 1.f, 0.f, 0.f, 0.f) ;
	}
	fprintf(sequence, "translate %f %f %f\n", xt, yt, zt) ;
      }
    }
    
  }

  // is frames/objects an integer?
  if ((nFrames%nFiles)!=0) {
    fprintf(stderr,
	    "Error: %s: number of frames, %d, is not divisible by "
	    "number of files, %d\n",name, nFrames,nFiles) ;
    return 1 ;
  }

  fprintf(sequence, "\n# End of data\n") ;
  fclose(sequence) ;
  fclose(frames) ;
  fclose(objects) ;
}
